unit UTComplex;

{
  Hier findet die komplette komplexe Arithmetik meines Programms statt.
  Die meisten Funktionen sind um den Overhead durch die Funktionsaufrufe zu
  vermeiden als volle Assemblerfunktionen geschrieben worden.

  Um Divisionen durch Null zu vermeiden ist Smallest eingefhrt worden.
  Vertretbare Werte sind hier Smallest<10^-25

  Um berlufe bei der Exponentialfunktion zu vermeiden gibt es einen MaxExp-
  Wert, der den Maximalen bzw. minimalen Exponenten bestimmmt.

  Das Winkelintervall bei der fpatan-Funktion ist glcklicherweise bei Intel-
  CPUs direkt auf das fr komplexe Zahlen bliche festgelegt. Auch die Unter-
  scheidung nach Quadranten wird getroffen.
  Quelle: http://www.ews.uiuc.edu/~cjiang/reference/vc107.htm
}

interface

uses SysUtils, math;

const e: extended = 2.71828182845904523536028747135266249;
//ich bezweifle, dass die vielen Stellen etwas ausmachen, aber es sieht nett aus
const OneHalf: extended = 0.5;
//wichtige Konstante, die der Coprozessor nicht kennt
const LnOneHalf: extended = 0.34657359027997265471;
//bentigt zur Bildung des natrlichen Logarithmus  ln2/2
const Smallest: extended = 1E-25;
  {Enthlt einen Wert der addiert wird um
   Divisionen durch Null zu vermeiden
   Standard: 10^-25}
const MaxExp: extended = 64;
  {Enthlt den maximalen und minimalen Wert
  der in die Exponentialfunktion eingesetzt wird
  Standard: 64 (meint e^-64<Wert<e^64)}

//kartesischer Typ
type
  TComplex = packed record
    real: extended;
    imag: extended;
  end;

//Polarformat, leider noch nicht vollstndig implementiert...
type
  TPolar = packed record
    r: extended;
    arg: extended;
  end;     

function MakeComplex(r, i: extended): TComplex;
function CAbs(C1: TComplex): extended; overload; //|Z|
function CModule(C1: TComplex): extended; //a^2+b^2 = |Z|^2
function CArg(C: TComplex): extended; //arg(Z) = phi

function CompToStr(C: TComplex): string; overload;
function CompToStr(Z: TComplex; prec: integer): string; overload;
function CCompare(A, B: TComplex): boolean;
//0 fr Gleichheit, 1 fr Ungleichheit

{
Die folgenden Funktionen sind fr den Parser im Format

  function(C1, C2: TComplex): TComplex

gehalten, ob sie nun zwei Argumente bentigen oder nicht spielt da keine Rolle.
Der Parser bentigt einen einheitlichen Funktionstyp um einen einheitlichen
Zeiger zu definieren.
}

function CAdd(var C1, C2: TComplex): TComplex; register;
function CSub(var C1, C2: TComplex): TComplex; register;     {RESULT := C1 - C2}
function CMul(var C1, C2: TComplex): TComplex; register;
function CDiv(var C1, C2: TComplex): TComplex; register;     {RESULT := C1 / C2}
function CPower(var C1, C2: TComplex): TComplex; register;   {RESULT := C1 ^ C2}
function CAbs(var C1, C2: TComplex): TComplex; register; overload;
function CMod(var C1, C2: TComplex): TComplex; register;
function CSqr(var C1, C2: TComplex): TComplex; register;
function CLn(var C1, C2: TComplex): TComplex; register;
function CExp(var C1, C2: TComplex): TComplex; register;
function CEul(var C1, C2: TComplex): TComplex; register;    {RESULT := e^(i*arg(C1))}
function CSin(var C1, C2: TComplex): TComplex; register;
function CCos(var C1, C2: TComplex): TComplex; register;
function CRep(var C1, C2: TComplex): TComplex; register;    {RESULT := C1.real}
function CImp(var C1, C2: TComplex): TComplex; register;    {RESULT := C1.imag}


function CRevSqrAdd(C1, C2: TComplex): TComplex;{C1:=C1^-2 + C2}
procedure CSqrAdd(var C1, C2: TComplex);        {C1:=C1^2 + C2}
//Funktionen fr Mandel- und NegBrot

function CompToPol(Z: TComplex): TPolar;
function PolToComp(P: TPolar): TComplex;
function PolPower(P: TPolar; exp: TComplex): TComplex;
//Unvollstndige Polarkoordinaten-Untersttzung
//Ursprnglich nur um andere Winkelintervalle zu berechnen

implementation

function MakeComplex(r, i: extended): TComplex;
begin
  result.real := r;
  result.imag := i;
end;

function CAbs(C1: TComplex): extended;
asm
  fld C1.real
  fmul ST(0),ST(0)
  fld C1.imag
  fmul ST(0),ST(0)
  fadd
  fsqrt
  fstp result
  fwait
end;

function CModule(C1: TComplex): extended;
asm
  fld C1.real
  fmul ST(0),ST(0)
  fld C1.imag
  fmul ST(0),ST(0)
  fadd
  fstp result
  fwait
end;

function CArg(C: TComplex): extended;
asm
  fld C.imag
  fld C.real
  fpatan
  fstp Result
  fwait
end;

function CompToStr(C: TComplex): string;
begin
  result := FloatToStr(C.real) + ' ' + FloatToStr(C.imag);
end;

function CompToStr(Z: TComplex; prec: integer): string;
begin
  Result := Format('%.' + IntToStr(prec) + 'f + i*%.' + IntToStr(prec) + 'f',
    [Z.real, Z.imag]);
end;

function CCompare(A, B: TComplex): boolean;
begin
  if (A.real = B.real) and (A.imag = B.imag) then Result := true
  else Result := false;
end;

//##############################################################################
//-------------------------Parser Funktionen------------------------------------
//##############################################################################

{ST(0)-ST(7) sind die  Register im Coprozessor-Stack.
Es ist jeweils die Position der Variablen im Stack
NACH Ausfhrung des Befehls vermerkt}

function CAdd(var C1, C2: TComplex): TComplex;
asm                {ST(0)       ST(1)   }
  fld C1.real      {C1.real             }
  fld C2.real      {C2.real     C1.real }
  fadd             {result.real         }
  fstp result.real {                    }
  fld C1.imag      {C1.imag             }
  fld C2.imag      {C2.imag     C1.imag }
  fadd             {result.imag}
  fstp result.imag
  fwait  //wartet auf prozessor
end;

function CSub(var C1, C2: TComplex): TComplex; //RESULT := C1 - C2 !!!!!
asm                {ST(0)       ST(1)   }
  fld C1.real      {C1.real             }
  fld C2.real      {C2.real     C1.real }
  fsub             {result.real         }
  fstp result.real {                    }
  fld C1.imag      {C1.imag             }
  fld C2.imag      {C2.imag     C1.imag }
  fsub             {result.imag}
  fstp result.imag
  fwait  //wartet auf prozessor
end;

function CMul(var C1, C2: TComplex): TComplex;
asm                 {ST(0)      ST(1)     ST(2)    ST(3)  }
  fld C2.real       {C2.real                              }
  fld C2.imag       {C2.imag    C2.real                   }
  fld C1.real       {C1.real    C2.imag   C2.real         }

  fmul ST(0), ST(2) {C1.real^2  C2.imag   C2.real         }
  fld C1.imag       {C1.imag    C1.real^2 C2.imag  C2.real}
  fmul ST(0), ST(2) {C1.imag^2  C1.real^2 C2.imag  C2.real}
  fsub              {result.r   C2.imag   C2.real         }
  fstp result.real  {C2.imag    C2.real                   }

  fld C1.real       {C1.real    C2.imag   C2.real         }
  fmul              {C1.r*C2.i  C2.real                   }
  fld C1.imag       {C1.imag    C1.r*C2.i C2.real         }
  fmul ST(0), ST(2) {C1.i*C2.r  C1.r*C2.i C2.real         }
  fadd              {result.i   C2.real                   }
  fstp result.imag  {C2.real                              }
  fstp st(0)        {Stack sollte bei reinen asm-Funktionen}
                    {am Ende geleert sein}
  fwait
end;

function CDiv(var C1, C2: TComplex): TComplex;
asm                 {ST(0)      ST(1)      ST(2)     ST(3)     ST(4)  }
  fld C2.real       {C2.real                                          }
  fmul ST(0),ST(0)  {C2.real^2                                        }
  fld C2.imag       {C2.imag    C2.real^2                             }
  fmul ST(0),ST(0)  {C2.imag^2  C2.real^2                             }
  fld  Smallest     {Smallest   C2.imag^2  C2.real^2                  }
  fadd              {S+C2.im^2  C2.real^2                             }
  fadd              {divisor                                          }

  fld C2.real       {C2.real    divisor                               }
  fld C2.imag       {C2.imag    C2.real    divisor                    }
  fld C1.real       {C1.real    C2.imag    C2.real   divisor          }
  fmul ST(0), ST(2) {C1.r*C2.r  C2.imag    C2.real   divisor          }
  fld C1.imag       {C1.imag    C1.r*C2.r  C2.imag   C2.real   divisor}
  fmul ST(0), ST(2) {C1.i*C2.i  C1.r*C2.r  C2.imag   C2.real   divisor}
  fadd              {a          C2.imag    C2.real   divisor          }
  fdiv ST(0), ST(3) {result.i   C2.imag    C2.real   divisor          }
  fstp result.real  {C2.imag    C2.real    divisor                    }

  fld C1.real       {C1.real    C2.imag    C2.real   divisor          }
  fmul              {C1.r*C2.i  C2.real    divisor                    }
  fld C1.imag       {C1.imag    C1.r*C2.i  C2.real   divisor          }
  fmul ST(0), ST(2) {C1.i*C2.r  C1.r*C2.i  C2.real   divisor          }
  fsubr             {b          C2.real    divisor                    }
  fdiv ST(0), ST(2) {b/div      C2.real    divisor                    }
  fstp result.imag  {C2.real    divisor                               }
  fstp st(0)        {divisor                                          }
  fstp st(0)        {Stack leeren                                     }

  fwait
end;

function CPower(var C1, C2: TComplex): TComplex;
asm                 {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
  fldln2            {ln2                               }
  fld C1.real       {C1.r   ln2                        }
  fmul ST(0),ST(0)  {C1.r^2 ln2                        }
  fld C1.imag       {C1.i   C1.r^2 ln2                 }
  fmul ST(0),ST(0)  {C1.i^2 C1.r^2 ln2                 }
  fld  Smallest     {Small  C1.i^2 C1.r^2 ln2          }
  fadd              {S+i1^2 C1.r^2 ln2                 }
  fadd              {modC1  ln2                        }
  fsqrt             {r      ln2                        }
  fyl2x             {ln(r)                             }

  fld C1.imag       {C1.i   ln(r)                      }
  fld C1.real       {C1.r   C1.i   ln(r)               }
  fpatan            {argC1  ln(r)                      }

  fld C2.real       {C2.r    argC1  ln(r)              }
  fmul ST(0), ST(2) {r2*lnr  argC1  ln(r)              }
  fld C2.imag       {C2.i    r2*lnr ln(r)              }
  fmul ST(0), ST(2) {i2*arg  r2*lnr argC1  ln(r)       }
  fsub              {a       argC1  ln(r)              }

  fld C2.real       {C2.r    a      argC1  ln(r)       }
  fmul ST(0), ST(2) {r2*arg  a      argC1  ln(r)       }
  fld C2.imag       {C2.i    r2*arg a      argC1  ln(r)}
  fmul ST(0), ST(4) {i2*lnr  r2*arg a      argC1  ln(r)}
  fadd              {b       a      argC1  ln(r)       }

  fsincos           {cos(b)  sin(b) a      arg1   ln(r)}
  fnstsw  ax        {Speichert das Status-Word des Coprozessors}
  sahf              {Bringt AX ins Flagregister des Prozessors}
  jpo    @@a1       {Parity-Flag des Prozessors entspricht nun}
  fstp   ST(0)      {Denormalisations-Flag des Coprozessors.}
  fld1              {Falls gesetzt berschreib cos(b) und sin(a)}
  fld1              {mit Einsen}
@@a1:
  fstp ST(4)        {sin(b)  a      argC1  cos(b)}
  fstp ST(4)        {a       argC1  cos(b) sin(b)}
 {Prft ob -MaxExp < a < MaxExp}
  fld MaxExp
  push eax          {sichere eax}
  fcom ST(1)        {vergleicht MaxExp mit a}
  fstsw ax          {bringt Statusword des Coprozessors}
  sahf              {ins Flagregister des Prozessors}
  jnc @@b1          {wenn Carry-Flag gesetzt (MaxExp>a)}
  fst ST(1)         {wenn nicht berschreibe a mit MaxExp}
  jmp @@c1
@@b1:
  fchs              {Aus MaxExp wird -MaxExp}
  fcom ST(1)        {analog zu oben}
  fstsw ax
  sahf
  jc @@c1
  fst ST(1)
@@c1:
  fstp ST(0)        {berschreibe MaxExp und fahre fort}
  pop eax           {hole eax zurck}

                    {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
  fldl2e            {lg2(e) a      argC1  cos(b) sin(b)}
  fmulp st(1),st(0) {a*lg2e argC1  cos(b) sin(b)       }
  fst   st(1)       {x      x      cos(b) sin(b)       }
  frndint           {rnd(x) x      cos(b) sin(b)       }
  fsub  st(1),st(0) {rnd(x) frc(x) cos(b) sin(b)       }
  fxch              {frc(x) rnd(x) cos(b) sin(b)       }
  f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
  fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
  fadd              {2^st   rnd(x) cos(b) sin(b)       }
  fscale            {2^x    rnd(x) cos(b) sin(b)       }
  fstp    st(1)     {exp(a) cos(b) sin(b)              }

  fmul ST(1), ST(0) {exp(a) Res.r  sin(b)              }
  fmulp ST(2),ST(0) {Res.r  Res.i                      }
  fstp Result.real  {Res.i                             }
  fstp Result.imag

  fwait
end;

function CAbs(var C1, C2: TComplex): TComplex;
asm
  fld C1.real
  fmul ST(0),ST(0)
  fld C1.imag
  fmul ST(0),ST(0)
  fadd
  fsqrt
  fstp result.real
  fldz
  fstp result.imag
  fwait
end;

function CMod(var C1, C2: TComplex): TComplex;
asm
  fld C1.real
  fmul ST(0),ST(0)
  fld C1.imag
  fmul ST(0),ST(0)
  fadd
  fstp result.real
  fldz
  fstp result.imag
  fwait
end;

function CSqr(var C1, C2: TComplex): TComplex;
asm                 {ST0        ST1     ST2     ST3     ST4     }
  fld C1.real       {C1.r                                       }
  fld C1.imag       {C1.i       C1.r                            }
  fld ST(1)         {C1.r       C1.i    C1.r                    }
  fmul ST(0),ST(0)  {C1.r^2     C1.i    C1.r                    }
  fld ST(1)         {C1.i       C1.r^2  C1.i    C1.r            }
  fmul ST(0),ST(0)  {C1.i^2     C1.r^2  C1.i    C1.r            }
  fsub              {result.r   C1.i    C1.r                    }
  fstp Result.real      {C1.i       C1.r                            }

  fmul              {C1.r*C1.i                                  }
  fld ST(0)         {C1.r*C1.i  C1.r*C1.i                       }
  fadd              {2*C1.r*C1.i                                }
  fstp Result.imag
  fwait
end;

function CExp(var C1, C2: TComplex): TComplex; //  V:=Exp(Z)
asm               {ST0      ST1     ST2     ST3     ST4}
    fld C1.imag   {C1.real                             }
    fsincos       {cos(a)   sin(a)                     }

    push eax      {siehe CPower}
    fnstsw ax     {prft auf Gltigkeit der Winkeloperation}
    sahf
    jpo @@a1
    fstp ST(0)
    fld1
    fld1
  @@a1:

    pop eax
    fld C1.real   {b        cos(a)  sin(a)             }

    fld MaxExp
    push eax          {sichere eax}
    fcom ST(1)        {vergleicht MaxExp mit a}
    fstsw ax          {bringt Statusword des Coprozessors}
    sahf              {ins Flagregister des Prozessors}
    jnc @@b1          {wenn Carry-Flag gesetzt (MaxExp>a)}
    fst ST(1)         {wenn nicht berschreibe a mit MaxExp}
    jmp @@c1
  @@b1:
    fchs              {Aus MaxExp wird -MaxExp}
    fcom ST(1)        {analog zu oben}
    fstsw ax
    sahf
    jc @@c1
    fst ST(1)
  @@c1:
    fstp ST(0)        {berschreibe MaxExp und fahre fort}
    pop eax           {hole eax zurck}

                      {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
    fldl2e            {lg2(e) b      cos(a) sin(a)       }
    fmul              {a*lg2e cos(a) sin(a)              }
    fld   ST(0)       {x      x      cos(b) sin(b)       }
    frndint           {rnd(x) x      cos(b) sin(b)       }
    fsub  ST(1),ST(0) {rnd(x) frc(x) cos(b) sin(b)       }
    fxch              {frc(x) rnd(x) cos(b) sin(b)       }
    f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
    fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
    fadd              {2^st   rnd(x) cos(b) sin(b)       }
    fscale            {2^x    rnd(x) cos(b) sin(b)       }
    fstp    ST(1)     {exp(a) cos(b) sin(b)              }

    fld ST(0)         {exp(a) exp(a) cos(b) sin(b)              }
    fmul ST(0), ST(2) {r.real exp(a) cos(b) sin(b)              }
    fstp Result.real  {exp(a) cos(b) sin(b)              }
    fmul ST(0), ST(2) {r.imag cos(b) sin(b)              }
    fstp Result.imag
    fstp ST(0)
    fstp ST(0)
end;

function CEul(var C1, C2: TComplex): TComplex; //Result:=e^i*arg(C1)
asm              {ST0       ST1      ST3}
  fld C1.imag    {C1.imag               }
  fld C1.real    {C1.real   C1.imag     }
  fpatan         {arg(C1)               }
  fsincos        {cos(arg)  sin(arg)    }
  fnstsw ax
  sahf
  jpo @@a1
  fstp ST(0)
  fld1
  fld1
@@a1:
  fstp Result.real
  fstp Result.imag
end;

function CLn(var C1, C2: TComplex): TComplex; //  Result:=Ln(C1)
asm                 {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
  fld  LnOneHalf    {ln2/2                             }
  fld C1.real       {C1.r   ln2/2                      }
  fmul ST(0),ST(0)  {C1.r^2 ln2/2                      }
  fld C1.imag       {C1.i   C1.r^2 ln2/2               }
  fmul ST(0),ST(0)  {C1.i^2 C1.r^2 ln2/2               }
  fld  Smallest     {Small  C1.i^2 C1.r^2 ln2/2        }
  fadd              {S+i1^2 C1.r^2 ln2/2               }
  fadd              {modC1  ln2/2                      }
  fyl2x             {ln(r)                             }
  fstp Result.real

  fld C1.imag
  fld C1.real
  fpatan
  fstp Result.imag
  fwait
end;

function CSin(var C1, C2: TComplex): TComplex;
asm               {ST0      ST1     ST2     ST3     ST4}
    fld C1.real   {C1.real                             }
    fsincos       {cos(a)   sin(a)                     }

    push eax
    fnstsw ax
    sahf
    jpo @@a1
    fstp ST(0)
    fld1
    fld1
  @@a1:

    pop eax
    fld C1.imag   {b        cos(a)  sin(a)             }

    fld MaxExp
    push eax          {sichere eax}
    fcom ST(1)        {vergleicht MaxExp mit a}
    fstsw ax          {bringt Statusword des Coprozessors}
    sahf              {ins Flagregister des Prozessors}
    jnc @@b1          {wenn Carry-Flag gesetzt (MaxExp>a)}
    fst ST(1)         {wenn nicht berschreibe a mit MaxExp}
    jmp @@c1
  @@b1:
    fchs              {Aus MaxExp wird -MaxExp}
    fcom ST(1)        {analog zu oben}
    fstsw ax
    sahf
    jc @@c1
    fst ST(1)
  @@c1:
    fstp ST(0)        {berschreibe MaxExp und fahre fort}
    pop eax           {hole eax zurck}

                      {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
    fldl2e            {lg2(e) b      cos(a) sin(a)       }
    fmul              {a*lg2e cos(a) sin(a)              }
    fld   ST(0)       {x      x      cos(b) sin(b)       }
    frndint           {rnd(x) x      cos(b) sin(b)       }
    fsub  ST(1),ST(0) {rnd(x) frc(x) cos(b) sin(b)       }
    fxch              {frc(x) rnd(x) cos(b) sin(b)       }
    f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
    fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
    fadd              {2^st   rnd(x) cos(b) sin(b)       }
    fscale            {2^x    rnd(x) cos(b) sin(b)       }
    fstp    ST(1)     {exp(a) cos(b) sin(b)              }

    fld onehalf       {0.5    exp(b)  cos(a)  sin(a)    }
    fdiv ST(0),ST(1)  {y      exp(b)  cos(a)  sin(a)  }
    fld onehalf       {0.5    y       exp(b)  cos(a)    sin(a)}
    fmulp ST(2),ST(0) {y      x       cos(a)  sin(a)}
    fld ST(1)         {x      y       x       cos(a)    sin(a)}
    fld ST(1)         {y      x       y       x         cos(a)    sin(a)}
    fadd              {x+y    y       x       cos(a)    sin(a)}
    fmul ST(0),ST(4)  {r.real y       x       cos(a)    sin(a)}
    fstp result.real  {y      x       cos(a)  sin(a)}
    fsubp ST(1),ST(0) {x-y    cos(a)  sin(a)}
    fmulp ST(1),ST(0) {r.imag cos(a)  sin(a)}
    fstp Result.imag  {sin(a)}
    fstp ST(0)
end;

function CCos(var C1, C2: TComplex): TComplex;
asm               {ST0      ST1     ST2     ST3     ST4}
    fld C1.real   {C1.real                             }
    fsincos       {cos(a)   sin(a)                     }

    push eax
    fnstsw ax
    sahf
    jpo @@a1
    fstp ST(0)
    fld1
    fld1
  @@a1:

    pop eax
    fld C1.imag   {b        cos(a)  sin(a)             }

    fld MaxExp
    push eax          {sichere eax}
    fcom ST(1)        {vergleicht MaxExp mit a}
    fstsw ax          {bringt Statusword des Coprozessors}
    sahf              {ins Flagregister des Prozessors}
    jnc @@b1          {wenn Carry-Flag gesetzt (MaxExp>a)}
    fst ST(1)         {wenn nicht berschreibe a mit MaxExp}
    jmp @@c1
  @@b1:
    fchs              {Aus MaxExp wird -MaxExp}
    fcom ST(1)        {analog zu oben}
    fstsw ax
    sahf
    jc @@c1
    fst ST(1)
  @@c1:
    fstp ST(0)        {berschreibe MaxExp und fahre fort}
    pop eax           {hole eax zurck}

                      {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
    fldl2e            {lg2(e) b      cos(a) sin(a)       }
    fmul              {a*lg2e cos(a) sin(a)              }
    fld   ST(0)       {x      x      cos(b) sin(b)       }
    frndint           {rnd(x) x      cos(b) sin(b)       }
    fsub  ST(1),ST(0) {rnd(x) frc(x) cos(b) sin(b)       }
    fxch              {frc(x) rnd(x) cos(b) sin(b)       }
    f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
    fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
    fadd              {2^st   rnd(x) cos(b) sin(b)       }
    fscale            {2^x    rnd(x) cos(b) sin(b)       }
    fstp    ST(1)     {exp(a) cos(b) sin(b)              }

    fld onehalf       {0.5    exp(b)  cos(a)  sin(a)    }
    fdiv ST(0),ST(1)  {y      exp(b)  cos(a)  sin(a)  }
    fld onehalf       {0.5    y       exp(b)  cos(a)    sin(a)}
    fmulp ST(2),ST(0) {y      x       cos(a)  sin(a)}
    fld ST(1)         {x      y       x       cos(a)    sin(a)}
    fld ST(1)         {y      x       y       x         cos(a)    sin(a)}
    fadd              {x+y    y       x       cos(a)    sin(a)}
    fmul ST(0),ST(3)  {r.real y       x       cos(a)    sin(a)}
    fstp result.real  {y      x       cos(a)  sin(a)}
    fsubp ST(1),ST(0) {x-y    cos(a)  sin(a)}
    fmul ST(0),ST(2)  {r.imag cos(a)  sin(a)}
    fchs
    fstp Result.imag  {cos(a) sin(a)}
    fstp ST(0)
    fstp ST(0)
end;

function CRep(var C1, C2: TComplex): TComplex; register;    {RESULT := C1.real}
asm
  fld C1.real
  fstp Result.real
  fldz
  fstp Result.imag
end;

function CImp(var C1, C2: TComplex): TComplex; register;    {RESULT := C1.imag}
asm
  fld C1.imag
  fstp Result.real
  fldz
  fstp Result.imag
end;

function CRevSqrAdd(C1, C2: TComplex): TComplex;
asm                 {ST0        ST1     ST2     ST3     ST4     }
  fld C1.real       {C1.r                                       }
  fld C1.imag       {C1.i       C1.r                            }
  fld ST(1)         {C1.r       C1.i    C1.r                    }
  fmul ST(0),ST(0)  {C1.r^2     C1.i    C1.r                    }
  fld ST(1)         {C1.i       C1.r^2  C1.i    C1.r            }
  fmul ST(0),ST(0)  {C1.i^2     C1.r^2  C1.i    C1.r            }
  fsub              {result.r   C1.i    C1.r                    }
  fstp ST(3)        {C1.i       C1.r    a                       }

  fmul              {C1.r*C1.i  a                               }
  fld ST(0)         {C1.r*C1.i  C1.r*C1.i a                     }
  fadd              {b          a                               }

  fld ST(0)         {b          b       a                       }
  fmul ST(0),ST(0)  {b^2        b       a                       }
  fld ST(2)         {a          b^2     b        a              }
  fmul ST(0),ST(0)  {a^2        b^2     b        a              }
  fadd              {a^2+b^2    b       a                       }
  fdiv ST(1),ST(0)  {a^2+b^2    re      a                       }
  fdiv ST(2),ST(0)  {a^2+b^2    re      im                      }
  fstp ST(0)        {re         im                              }
  fchs
  fld C2.imag       {C2.real    re      im                      }
  fadd              {res.r      im                              }
  fstp Result.imag  {im                                         }
  //fchs              {-im                                        }
  fld C2.real       {C2.imag    -im                             }
  fadd              {res.im                                     }
  fstp Result.real

  fwait
end;

procedure CSqrAdd(var C1, C2: TComplex);
asm                 {ST0        ST1     ST2     ST3     ST4     }
  fld C1.real       {C1.r                                       }
  fld C1.imag       {C1.i       C1.r                            }
  fld ST(1)         {C1.r       C1.i    C1.r                    }
  fmul ST(0),ST(0)  {C1.r^2     C1.i    C1.r                    }
  fld ST(1)         {C1.i       C1.r^2  C1.i    C1.r            }
  fmul ST(0),ST(0)  {C1.i^2     C1.r^2  C1.i    C1.r            }
  fsub              {result.r   C1.i    C1.r                    }
  fstp ST(3)        {C1.i       C1.r    a                       }

  fmul              {C1.r*C1.i  a                               }
  fld ST(0)         {C1.r*C1.i  C1.r*C1.i a                     }
  fadd              {b          a                               }

  fld ST(0)         {b          b       a                       }
  fmul ST(0),ST(0)  {b^2        b       a                       }
  fld ST(2)         {a          b^2     b        a              }
  fmul ST(0),ST(0)  {a^2        b^2     b        a              }
  fadd              {a^2+b^2    b       a                       }
  fdiv ST(1),ST(0)  {a^2+b^2    re      a                       }
  fdiv ST(2),ST(0)  {a^2+b^2    re      im                      }
  fstp ST(0)        {re         im                              }
  fld C2.real       {C2.real    re      im                      }
  fadd              {res.r      im                              }
  fstp C1.real
  fld C2.imag
  fadd
  fstp C1.imag

  fwait
end;

function CompToPol(Z: TComplex): TPolar;
begin
  Result.r := sqrt(Z.real * Z.real + Z.imag * Z.imag);
  Result.arg := arctan2(Z.imag, Z.real);
end;

function PolToComp(P: TPolar): TComplex;
begin
  Result.real := P.r * cos(P.arg);
  Result.imag := P.r * sin(P.arg);
end;

function PolPower(P: TPolar; exp: TComplex): TComplex;
var
  bet, bet2: extended;
begin
  if p.r <= 0 then
    p.r := 1E-25;
  bet := power(e, ln(p.r) * exp.real - P.arg * exp.imag + 1E-25);
  bet2 := exp.imag * ln(p.r) + exp.real * p.arg;
  result.real := bet * cos(bet2);
  result.imag := bet * sin(bet2);
end;


end.
