unit UOGL;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, GL, utpixel; //LCLType, ftfont, UTPixel, freetype;

type
    OGLTexture = packed record
      data: ^Longword;
      Width, Height: integer;
      ID: GluInt;
    end;

    OGLTextureTile = packed record
      data: ^Longword;
      ID: GluInt;
    end;

    OGLTiledTexture = packed record
      Tiles: array of OGLTextureTile;
      TileSize: integer;
      nmbrTilesX, nmbrTilesY: integer;
      Width, Height: integer;
    end;

    OGLVScrollBar = packed record
      Enabled, Visible: boolean;
      ButtonHeight: Glfloat;
      TopButton, LowerButton: Glfloat;
      Position, pageSize: Glfloat;
      roundEdge, shadow: Glfloat;
      Left: Glfloat;
      Width: Glfloat;
      ScrollState: integer;
      dragState: boolean;
      dragPos: GLfloat;
      scrRelation: Glfloat;
    end;

    OGLHScrollBar = packed record
      Enabled, Visible: boolean;
      ButtonWidth: Glfloat;
      LeftButton, RightButton: Glfloat;
      Position, pageSize: Glfloat;
      roundEdge: Glfloat;
      Top: Glfloat;
      Height: Glfloat;
      ScrollState: integer;
      scrRelation: Glfloat;
      dragPos: Glfloat;
      dragState: boolean;
    end;

const

    EdgeX: array[0..10] of GlFloat =
    ( 1.00 , 0.90 , 0.80 , 0.70 , 0.60 , 0.50 , 0.40 , 0.30 , 0.20 , 0.10 , 0.00);
    EdgeY: array[0..10] of GlFloat =
    ( 0.00 , 0.01 , 0.02 , 0.05 , 0.08 , 0.13 , 0.20 , 0.29 , 0.4 , 0.66 , 1.00);
    //( 0.00 , 0.44 , 0.60 , 0.71 , 0.80 , 0.87 , 0.92 , 0.95 , 0.98 , 0.99 , 1.00);

procedure GenerateOGLTiledTexture(var T: OGLTiledTexture; W, H: integer);
procedure FillQuad(var T: OGLTiledTexture; P: TPixel; X,Y,S: integer); inline;
procedure SetPixel(var T: OGLTiledTexture; P: TPixel; X,Y: integer); inline;
procedure UpdateTexture(T: OGLTiledTexture);
procedure DrawVScrollBar(var S: OGLVScrollBar);
procedure DrawHScrollBar(var S: OGLHScrollBar);
function ScrollY(var S: OGLVScrollBar; A: Double): boolean;
function ScrollH(var S: OGLHScrollBar; A: Double): boolean;
procedure LineF(X1, Y1, X2, Y2: Single; P: TPixel; var T: OGLTiledTexture);
procedure SetPixelF(X,Y: Single; P: TPixel; var T: OGLTiledTexture);
procedure LineH(X1, Y1, l, th: integer; P: TPixel; var T: OGLTiledTexture);
procedure LineV(X1, Y1, l, th: integer; P: TPixel; var T: OGLTiledTexture);

implementation
uses UTFractalGenerator;

procedure SetPixelF(X,Y: Single; P: TPixel; var T: OGLTiledTexture);
var
  iX, iY: integer;
  fracX,fracY: Single;
  t0: integer;
  p0,p1,p2,p3: integer;
  a0,a1,a2,a3: Double;
begin
  iX := round(X);
  iY := round(Y);
  t0 := iX div 512 + (iY div 512) * T.nmbrTilesX;
  p0 := ix mod 512 + (iY mod 512) * 512;

  //Festlegung zwischen welchen vier Pixeln antialiast wird
  if X < iX then
  begin
    fracX := 1 - frac(X);
    p1 := p0-1;
    if Y < iY then
    begin
      fracY := 1 - frac(Y);
      p2 := p0+T.TileSize;
      p3 := p0+T.TileSize - 1;
    end else begin
      fracY := frac(Y);
      p2 := p0-T.TileSize;
      p3 := p0-T.TileSize - 1;
    end;
  end else begin
    fracX := frac(X);
    p1 := p0+1;
    if Y < iY then
    begin
      fracY := 1 - frac(Y);
      p2 := p0+T.TileSize;
      p3 := p0+T.TileSize + 1;
    end else begin
      fracY := frac(Y);
      p2 := p0 - T.TileSize;
      p3 := p0 - T.TileSize + 1;
    end;
  end;

  a0 := sqrt(sqr(fracX)+sqr(fracY));
  a1 := sqrt(sqr(fracX-1)+sqr(fracY));
  a2 := sqrt(sqr(fracX-1)+sqr(fracY-1))*0.666;
  a3 := sqrt(sqr(fracX)+sqr(fracY-1));

  T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
  T.Tiles[t0].data[p1] := BlendPixel(T.Tiles[t0].data[p1], P, a1);
  T.Tiles[t0].data[p2] := BlendPixel(T.Tiles[t0].data[p2], P, a2);
  T.Tiles[t0].data[p3] := BlendPixel(T.Tiles[t0].data[p3], P, a3);
end;

procedure Swap(var a,b: integer);
var
  tmp: integer;
begin
  tmp := a;
  a := b;
  b := tmp;
end;

procedure Line(X0,Y0,X1,Y1: integer; P: TPixel; var T: OGLTiledTexture);
var
  steep: boolean;
  deltax, deltay: integer;
  error, deltaerr: Single;
  y, x, ystep, t0 : integer;
begin
  steep := abs(y1 - y0) > abs(x1 - x0);
  if steep then
  begin
    swap(x0, y0);
    swap(x1, y1);
  end;
  if x0 > x1 then
  begin
    swap(x0, x1);
    swap(y0, y1);
  end;
  deltax := x1 - x0;
  deltay := abs(y1 - y0);
  error := 0;
  deltaerr := deltay / deltax;
  y := y0;
  if y0 < y1 then ystep := 1 else ystep := -1;
  for x := x0 to x1 do
  begin
    if steep then SetPixel(T,P,Y,X) else SetPixel(T,P,X,Y);
    error := error + deltaerr;
    if error >= 0.5 then
    begin
      y := y + ystep;
      error := error - 1.0;
    end;
  end;
end;

procedure LineV(X1, Y1, l, th: integer; P: TPixel; var T: OGLTiledTexture);
var
  i,j,thl,thh: integer;
begin
  if th > 1 then
  begin
    thl := th div 2;
    if X1 - thl < 0 then thl := X1;
    thh := th - thl;
  end else begin
    thl:=0;
    thh:=0;
  end;

  for j := X1 - thl to X1 + thh do
  begin
    for i:=Y1 to Y1 + l do
      SetPixel(T,P,j,i);
  end;
end;

procedure LineH(X1, Y1, l, th: integer; P: TPixel; var T: OGLTiledTexture);
var
  i,j,thl,thh: integer;
begin
  if th > 1 then
  begin
    thl := th div 2;
    if Y1 - thl < 0 then thl := Y1;
    thh := th - thl;
  end else begin
    thl:=0;
    thh:=0;
  end;
  for j := Y1 - thl to Y1 + thh do
  begin
    for i:=X1 to X1 + l do
      SetPixel(T,P,i,j);
  end;
end;

procedure LineF(X1, Y1, X2, Y2: Single; P: TPixel; var T: OGLTiledTexture);
var
  dx, dy, gradient, a0, intery: Single;
  tmp, yend, xgap: Single;
  xend, xpxl1, ypxl1, xpxl2, ypxl2: integer;
  p0, t0, x: integer;
begin
  dx := x2 - x1;
  dy := y2 - y1;

if abs(dx) > abs(dy) then
begin
  if abs(dx) > abs(dy) then
  //handle "horizontal" lines
  if x2 < x1 then
  begin
    tmp := x1;
    x1 := x2;
    x2 := tmp;
    tmp := y1;
    y1 := y2;
    y2 := tmp;
  end;
  gradient := dy / dx;

     // handle first endpoint
  xend := round(x1);
  yend := y1 + gradient * (xend - x1);
  xgap := 1 - frac(x1 + 0.5);
  xpxl1 := xend;  // this will be used in the main loop
  ypxl1 := trunc(yend);

  t0 := xpxl1 div 512 + (ypxl1 div 512) * T.nmbrTilesX;
  p0 := xpxl1 mod 512 + (ypxl1 mod 512) * 512;
  a0 := (1-frac(yend)) * xgap;
  T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
  a0 := frac(yend) * xgap;
  T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);
  intery := yend + gradient;

  xend := round(x2);
  yend := y2 + gradient * (xend - x2);
  xgap := frac(x2 + 0.5);
  xpxl2 := xend;
  ypxl2 := trunc(yend);
  t0 := xpxl2 div 512 + (ypxl2 div 512) * T.nmbrTilesX;
  p0 := xpxl2 mod 512 + (ypxl2 mod 512) * 512;
  a0 := (1-frac(yend)) * xgap;
  T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
  a0 := frac(yend) * xgap;
  T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);

  for x := xpxl1 + 1 to xpxl2 - 1 do
  begin
    t0 := x div 512 + (trunc(intery) div 512) * T.nmbrTilesX;
    p0 := x mod 512 + (trunc(intery) mod 512) * 512;
    a0 := 1-frac(intery);
    T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
    a0 := frac(intery);
    T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);
    intery := intery + gradient;
  end;
end else begin
  tmp := y1;
  y1 := x1;
  x1 := tmp;
  tmp := y2;
  y2 := x2;
  x2 := tmp;
  dx := x2 - x1;
  dy := y2 - y1;
  if abs(dx) > abs(dy) then
  //handle "horizontal" lines
  if x2 < x1 then
  begin
    tmp := x1;
    x1 := x2;
    x2 := tmp;
    tmp := y1;
    y1 := y2;
    y2 := tmp;
  end;
  gradient := dy / dx;

     // handle first endpoint
  xend := round(x1);
  yend := y1 + gradient * (xend - x1);
  xgap := 1 - frac(x1 + 0.5);
  xpxl1 := xend;  // this will be used in the main loop
  ypxl1 := trunc(yend);

  t0 := ypxl1 div 512 + (xpxl1 div 512) * T.nmbrTilesX;
  p0 := ypxl1 mod 512 + (xpxl1 mod 512) * 512;
  a0 := (1-frac(yend)) * xgap;
  T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
  a0 := frac(yend) * xgap;
  T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);
  intery := yend + gradient;

  xend := round(x2);
  yend := y2 + gradient * (xend - x2);
  xgap := frac(x2 + 0.5);
  xpxl2 := xend;
  ypxl2 := trunc(yend);
  t0 := ypxl2 div 512 + (xpxl2 div 512) * T.nmbrTilesX;
  p0 := ypxl2 mod 512 + (xpxl2 mod 512) * 512;
  a0 := (1-frac(yend)) * xgap;
  T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
  a0 := frac(yend) * xgap;
  T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);

  for x := xpxl1 + 1 to xpxl2 - 1 do
  begin
    t0 := trunc(intery) div 512 + (x div 512) * T.nmbrTilesX;
    p0 := trunc(intery) mod 512 + (x mod 512) * 512;
    a0 := 1-frac(intery);
    T.Tiles[t0].data[p0] := BlendPixel(T.Tiles[t0].data[p0], P, a0);
    a0 := frac(intery);
    T.Tiles[t0].data[p0-512] := BlendPixel(T.Tiles[t0].data[p0-512], P, a0);
    intery := intery + gradient;
  end;
end;
end;

procedure GenerateOGLTiledTexture(var T: OGLTiledTexture; W, H: integer);
var
  i: integer;
begin
  T.TileSize := 512;
  //Größe der Kacheln, muss "Power of Two" sein

  for i := 0 to High(T.Tiles) do
    if Assigned(T.Tiles[i].data) then
      System.FreeMem(T.Tiles[i].data, sqr(T.TileSize)*4);
      //falls Textur nicht leer

  T.Width := W;
  T.Height := H;
  T.nmbrTilesX := T.Width div T.TileSize + 1;
  T.nmbrTilesY := T.Height div T.TileSize + 1;

  SetLength(T.Tiles, T.nmbrTilesX * T.nmbrTilesY);

  glEnable(GL_TEXTURE_2D);

  for i := 0 to High(T.Tiles) do
  begin
    GetMem(T.Tiles[i].data, sqr(T.TileSize) * 4);
    FillDword(T.Tiles[i].data^, sqr(T.TileSize), TexBackground);
    glGenTextures(1, @T.Tiles[i].ID);
    glBindTexture(GL_TEXTURE_2D, T.Tiles[i].ID);

    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    glTexImage2D(GL_TEXTURE_2D, 0, 4, T.TileSize, T.TileSize,
                           0, GL_RGBA, GL_UNSIGNED_BYTE, T.Tiles[i].data);
  end;
end;

procedure FillQuad(var T: OGLTiledTexture; P: TPixel; X,Y,S: integer); inline;
var
   i,k: integer;
   d: PPixel;
begin
for i := X to X + S do
  for k := Y to Y + S do
    SetPixel(T, P, i, k);
{     d:=T.Tiles[0].data+X+Y*T.Width;

     for i:=S-1 downto 0 do
     begin
          for k:=S-1 downto 0 do
          begin
               d[0]:=P;
               inc(d);
          end;
          inc(d, T.Width-S);
     end;}
end;

procedure SetPixel(var T: OGLTiledTexture; P: TPixel; X,Y: integer); inline;
begin
     T.Tiles[X div 512 + (Y div 512) * T.nmbrTilesX
      ].data[ X mod 512 + (Y mod 512) * 512] := P;
end;

procedure UpdateTexture(T: OGLTiledTexture);
var
  i: integer;
begin
  for i := 0 to High(T.Tiles) do
  begin
     glBindTexture(GL_TEXTURE_2D, T.Tiles[i].ID);
     glTexImage2D(GL_TEXTURE_2D, 0, 4, T.TileSize, T.TileSize,
                           0, GL_RGBA, GL_UNSIGNED_BYTE, T.Tiles[i].data);
  end;
end;

function ScrollH(var S: OGLHScrollBar; A: Double): boolean;
var
  scr: Glfloat;
begin
     Result:=false;
     with S do
     begin
          if not Enabled then Exit;
          scr:=A;
          if scr<0 then
          begin
               if Position=0 then Exit
               else if Position+scr<=0 then
               begin
                    ScrollState:=1;
                    scr:=-Position;
                    Result:=true;
               end else Result:=true;
          end else begin
              if Position+pageSize=RightButton-LeftButton-ButtonWidth-1 then Exit
              else if Position+pageSize+scr>=RightButton-LeftButton-ButtonWidth-1 then
              begin
                   ScrollState:=-1;
                   scr:=abs( (Position+pageSize) - (RightButton-LeftButton-ButtonWidth-1));
                   Result:=true;
              end else Result:=true;
          end;
          Position:=Position+scr;
          FG.scr_posX:=FG.scr_posX-scr/scrRelation;
     end;
end;

function ScrollY(var S: OGLVScrollBar; A: Double): boolean;
var
  scr: Glfloat;
begin
     Result:=false;
     with S do
     begin
          if not Enabled then Exit;
          scr:=A;
          if scr<0 then
          begin
               if Position=0 then Exit
               else if Position+scr<=0 then
               begin
                    ScrollState:=1;
                    scr:=-Position;
                    Result:=true;
               end else
               begin
                Result:=true;
                ScrollState:=0;
               end;
          end else begin
              if Position+pageSize=LowerButton-TopButton-ButtonHeight-1 then Exit
              else if Position+pageSize+scr>=LowerButton-TopButton-ButtonHeight-1 then
              begin
                   ScrollState:=-1;
                   scr:=abs( (Position+pageSize) - (LowerButton-TopButton-ButtonHeight-1));
                   Result:=true;
              end else begin
                ScrollState:=0;
                Result:=true;
              end;
          end;
          Position:=Position+scr;
          FG.scr_posY:=FG.scr_posY-scr/scrRelation;
     end;
end;

procedure DrawHScrollBar(var S: OGLHScrollBar);
var
   i: integer;
begin
     glColor3f(0.75,0.75,0.75);
     //Scrollhintergrund
     glBegin(GL_QUADS);
        glVertex3f(S.LeftButton+S.ButtonWidth, S.Top+S.Height, -1);
        glVertex3f(S.RightButton, S.Top+S.Height, -1);
        glVertex3f(S.RightButton, S.Top, -1);
        glVertex3f(S.LeftButton+S.ButtonWidth, S.Top, -1);
     glEnd;

     glColor3f(0.6,0.6,0.6);
     glLineWidth(1.15);
     //linker Scrollbutton-Rand
     glBegin(GL_LINE_LOOP);
       glVertex3f(S.LeftButton+S.ButtonWidth, S.Top+S.Height, -1);
       glVertex3f(S.LeftButton+S.ButtonWidth, S.Top, -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.LeftButton + S.roundEdge * EdgeX[i],
              S.Top + S.roundEdge * EdgeY[i], -1);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.LeftButton + S.roundEdge * EdgeY[i],
              S.Top + S.Height - S.roundEdge * EdgeX[i], -1);
     glEnd;
     //rechter Scrollbutton-Rand
     glBegin(GL_LINE_LOOP);
       glVertex3f(S.RightButton, S.Top, -1);
       glVertex3f(S.RightButton, S.Top+S.Height, -1);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.RightButton + S.ButtonWidth - S.roundEdge * EdgeX[i],
              S.Top + S.Height - S.roundEdge * EdgeY[i], -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.RightButton + S.ButtonWidth - S.roundEdge * EdgeY[i],
              S.Top + S.roundEdge * EdgeX[i], -1);
     glEnd;

     glColor3f(0.9,0.9,0.9);
     //linker Scrollbutton
     glBegin(GL_TRIANGLE_FAN);
       glVertex3f(S.LeftButton+S.ButtonWidth, S.Top+S.Height, -1);
       glColor3f(1,1,1);
       glVertex3f(S.LeftButton+S.ButtonWidth, S.Top, -1);
       //Rundung links oben
       for i:=0 to 10 do
         glVertex3f(S.LeftButton + S.roundEdge * EdgeX[i],
              S.Top + S.roundEdge * EdgeY[i], -1);
       //Rundung links unten
       for i:=0 to 10 do
         glVertex3f(S.LeftButton + S.roundEdge * EdgeY[i],
              S.Top + S.Height - S.roundEdge * EdgeX[i], -1);
       glColor3f(0.9,0.9,0.9);
       glVertex3f(S.LeftButton+S.ButtonWidth, S.Top+S.Height, -1);
     glEnd;
     //rechter Scrollbutton
     glBegin(GL_TRIANGLE_FAN);
       glColor3f(1,1,1);
       glVertex3f(S.RightButton+S.ButtonWidth/2, S.Top, -1);
       glVertex3f(S.RightButton, S.Top, -1);
       glVertex3f(S.RightButton, S.Top+S.Height, -1);
       glColor3f(0.9,0.9,0.9);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.RightButton + S.ButtonWidth - S.roundEdge * EdgeX[i],
              S.Top + S.Height - S.roundEdge * EdgeY[i], -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.RightButton + S.ButtonWidth - S.roundEdge * EdgeY[i],
              S.Top + S.roundEdge * EdgeX[i], -1);
       glVertex3f(S.RightButton+S.ButtonWidth/2, S.Top, -1);
     glEnd;

    //Scroll-Pfeile
    glColor3f(0.75,0.75,0.75);
    glBegin(GL_TRIANGLES);
    //links
       glVertex3f(S.LeftButton+S.ButtonWidth*0.75, S.Top+S.Height*0.25, -1);
       glVertex3f(S.LeftButton+S.ButtonWidth*0.25, S.Top+S.Height*0.5, -1);
       glVertex3f(S.LeftButton+S.ButtonWidth*0.75, S.Top+S.Height*0.75, -1);
    //rechts
       glVertex3f(S.RightButton+S.ButtonWidth*0.25, S.Top+S.Height*0.25, -1);
       glVertex3f(S.RightButton+S.ButtonWidth*0.25, S.Top+S.Height*0.75, -1);
       glVertex3f(S.RightButton+S.ButtonWidth*0.75, S.Top+S.Height*0.5, -1);
    glEnd;

    //Scrollbar
    glColor3f(0.261, 0.406, 0.607);
    glLineWidth(1.15);
    glBegin(GL_LINE_LOOP);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth, S.Top+S.Height, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top+S.Height, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth, S.Top, -1);
    glEnd;
    glBegin(GL_TRIANGLE_FAN);
      glColor3f(0.561, 0.706, 0.907);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top+S.Height/2, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth, S.Top, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth, S.Top+S.Height/2, -1);
      glColor3f(0.481, 0.627, 0.807);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth, S.Top+S.Height, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top+S.Height, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize, S.Top+S.Height/2, -1);
    glEnd;

    glColor3f(0.261, 0.406, 0.607);
    glLineWidth(1.15);
    glBegin(GL_LINES);
      glVertex3f(S.LeftButton+S.ButtonWidth+S.Position+S.pageSize*0.5-5, S.Top+S.Height*0.33, -1);
      glVertex3f(S.LeftButton+S.ButtonWidth+S.Position+S.pageSize*0.5-5, S.Top+S.Height*0.66, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize*0.5, S.Top+S.Height*0.33, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize*0.5, S.Top+S.Height*0.66, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize*0.5+5, S.Top+S.Height*0.33, -1);
      glVertex3f(S.LeftButton+S.Position+S.ButtonWidth+S.pageSize*0.5+5, S.Top+S.Height*0.66, -1);
    glEnd;


end;

procedure DrawVScrollBar(var S: OGLVScrollBar);
var
   i: integer;
begin
     glColor3f(0.75,0.75,0.75);
     //Scrollhintergrund
     glBegin(GL_QUADS);
        glVertex3f(S.Left, S.TopButton+S.ButtonHeight, -1);
        glVertex3f(S.Left, S.LowerButton, -1);
        glVertex3f(S.Left+S.Width, S.LowerButton, -1);
        glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight, -1);
     glEnd;

     glColor3f(0.6,0.6,0.6);
     glLineWidth(1.15);
     //oberer Scrollbutton-Rand
     glBegin(GL_LINE_LOOP);
       glVertex3f(S.Left, S.TopButton+S.ButtonHeight, -1);
       glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight, -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.Left + S.Width - S.roundEdge * EdgeY[i],
              S.TopButton + S.roundEdge * EdgeX[i], -1);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.Left + S.roundEdge * EdgeX[i],
              S.TopButton + S.roundEdge * EdgeY[i], -1);
     glEnd;

     glColor3f(0.9,0.9,0.9);
     //oberer Scrollbutton
     glBegin(GL_TRIANGLE_FAN);
       glVertex3f(S.Left, S.TopButton+S.ButtonHeight, -1);
       glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight, -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.Left + S.Width - S.roundEdge * EdgeY[i],
              S.TopButton + S.roundEdge * EdgeX[i], -1);
       //Rundung links
       glColor3f(1,1,1);
       for i:=0 to 9 do
         glVertex3f(S.Left + S.roundEdge * EdgeX[i],
              S.TopButton + S.roundEdge * EdgeY[i], -1);
        glColor3f(0.9,0.9,0.9);
       glVertex3f(S.Left,
              S.TopButton + S.roundEdge, -1);
       glVertex3f(S.Left, S.TopButton+S.ButtonHeight, -1);
     glEnd;

     glColor3f(0.6,0.6,0.6);
     glLineWidth(1.15);
     //unterer Scrollbutton-Rand
     glBegin(GL_LINE_LOOP);
       glVertex3f(S.Left, S.LowerButton, -1);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.Left + S.roundEdge * EdgeY[i],
              S.LowerButton + S.ButtonHeight - S.roundEdge * EdgeX[i], -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.Left + S.Width - S.roundEdge * EdgeX[i],
              S.LowerButton + S.ButtonHeight - S.roundEdge * EdgeY[i], -1);
       glVertex3f(S.Left + S.Width, S.LowerButton, -1);
     glEnd;
     glColor3f(0.9,0.9,0.9);
     //unterer Scrollbutton
     glBegin(GL_TRIANGLE_FAN);
       glVertex3f(S.Left, S.LowerButton + S.ButtonHeight*0.75, -1);
       //Rundung links
       for i:=0 to 10 do
         glVertex3f(S.Left + S.roundEdge * EdgeY[i],
              S.LowerButton + S.ButtonHeight - S.roundEdge * EdgeX[i], -1);
       //Rundung rechts
       for i:=0 to 10 do
         glVertex3f(S.Left + S.Width - S.roundEdge * EdgeX[i],
              S.LowerButton + S.ButtonHeight - S.roundEdge * EdgeY[i], -1);
       glVertex3f(S.Left + S.Width, S.LowerButton, -1);
       glColor3f(1,1,1);
       glVertex3f(S.Left + S.RoundEdge, S.LowerButton, -1);
       glVertex3f(S.Left, S.LowerButton, -1);
       glVertex3f(S.Left, S.LowerButton + S.ButtonHeight*0.75, -1);
     glEnd;
    //Scroll-Pfeile
    glColor3f(0.75,0.75,0.75);
    glBegin(GL_TRIANGLES);
    //oben
       glVertex3f(S.Left+S.Width*0.50, S.TopButton+S.ButtonHeight*0.25, -1);
       glVertex3f(S.Left+S.Width*0.25, S.TopButton+S.ButtonHeight*0.75, -1);
       glVertex3f(S.Left+S.Width*0.75, S.TopButton+S.ButtonHeight*0.75, -1);
    //unten
       glVertex3f(S.Left+S.Width*0.25, S.LowerButton+S.ButtonHeight*0.25, -1);
       glVertex3f(S.Left+S.Width*0.50, S.LowerButton+S.ButtonHeight*0.75, -1);
       glVertex3f(S.Left+S.Width*0.75, S.LowerButton+S.ButtonHeight*0.25, -1);
    glEnd;
    //Scrollbar
    glColor3f(0.261, 0.406, 0.607);
    glLineWidth(1.15);
    glBegin(GL_LINE_LOOP);
      glVertex3f(S.Left, S.TopButton+S.ButtonHeight+S.Position, -1);
      glVertex3f(S.Left, S.TopButton+S.ButtonHeight+S.Position+S.pageSize, -1);
      glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight+S.Position+S.pageSize, -1);
      glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight+S.Position, -1);
    glEnd;
    glBegin(GL_TRIANGLE_FAN);
      glColor3f(0.561, 0.706, 0.907);
      glVertex3f(S.Left+S.Width*0.4, S.TopButton+S.ButtonHeight+S.Position, -1);
      glVertex3f(S.Left, S.TopButton+S.ButtonHeight+S.Position, -1);
      glVertex3f(S.Left, S.TopButton+S.ButtonHeight+S.Position+S.pageSize, -1);
      glColor3f(0.481, 0.627, 0.807);
      glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight+S.Position+S.pageSize, -1);
      glVertex3f(S.Left+S.Width, S.TopButton+S.ButtonHeight+S.Position, -1);
    glEnd;

    glColor3f(0.261, 0.406, 0.607);
    glLineWidth(1.15);
    glBegin(GL_LINES);
      glVertex3f(S.Left+S.Width*0.33, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5-5, -1);
      glVertex3f(S.Left+S.Width*0.66, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5-5, -1);
      glVertex3f(S.Left+S.Width*0.33, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5, -1);
      glVertex3f(S.Left+S.Width*0.66, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5, -1);
      glVertex3f(S.Left+S.Width*0.33, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5+5, -1);
      glVertex3f(S.Left+S.Width*0.66, S.TopButton+S.ButtonHeight+S.Position+S.pageSize*0.5+5, -1);

    glEnd;

end;

end.

