unit UTComplex;

interface
{$I cfg_symbols.inc}

uses SysUtils,
     UMisc
{$ifndef UseAsm} ,Math {$endif}
     ;

const e: Double = 2.7182818284590452;
const OneHalf: Double = 0.5;
//wichtige Konstante, die der Coprozessor nicht kennt
const LnOneHalf: Double = 0.34657359027997265471;
//benötigt zur Bildung des natürlichen Logarithmus ln2/2
const Smallest: Double = 1E-25;
  {Enthält einen Wert der addiert wird um
   Divisionen durch Null zu vermeiden
   Standard: 10^-25}
const MaxExp: Double = 64;
  {Enthält den maximalen und minimalen Wert
  der in die Exponentialfunktion eingesetzt wird
  Standard: 64 (meint e^-64<Wert<e^64)}

//kartesischer Typ
type
  TComplex = packed record
    real: Double;
    imag: Double;
  end;

operator := (r: real) C: TComplex; inline;
operator + (C1, C2 : TComplex) C : Tcomplex;
operator - (C1, C2 : TComplex) C : Tcomplex;
operator * (C1, C2 : TComplex) C : Tcomplex;
operator * (C1: TComplex; r2: Double) C : Tcomplex;
operator * (r1: Double; C2: TComplex) C : Tcomplex;
operator / (C1, C2 : TComplex) C : Tcomplex;
operator ** (C1, C2 : TComplex) C : Tcomplex;
operator ** (C1: TComplex; r2: Double) C : Tcomplex;
operator = (C1, C2 : TComplex) b : boolean; inline;

function CSqr(C1: TComplex): TComplex;
function CAbs(C1: TComplex): Double;
function CLn(C1: TComplex): TComplex;
function CExp(C1: TComplex): TComplex;
function MakeComplex(r, i: Double): TComplex; inline;
function CModule(C1: TComplex): Double; inline;
function CArg(C: TComplex): Double;
function CompToStr(C: TComplex; prec: integer = 5): string;
function CDouble(C1: TComplex): TComplex;
function CConj(C: TComplex): TComplex;

procedure CSqrAdd(var C1, C2: TComplex);        {C1:=C1^2 + C2}

implementation

operator := (r: real) C: TComplex; inline;
begin
     Result.real:=r;
     Result.imag:=0.0;
end;

operator + (C1, C2 : TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm                            {ST(0)       ST(1)   }
  fld qword ptr [C1]           {C1.real             }
  fld qword ptr [C2]           {C2.real     C1.real }
  faddp ST(1), ST(0)           {result.real         }
  fstp qword ptr [result]      {                    }
  fld qword ptr [C1+8]         {C1.imag             }
  fld qword ptr [C2+8]         {C2.imag     C1.imag }
  faddp ST(1), ST(0)           {result.imag}
  fstp qword ptr [result+8]
end;
{$else}
inline;
begin
     Result.real:=C1.real+C2.real;
     Result.imag:=C1.imag+C2.imag;
end;
{$endif}

operator - (C1, C2 : TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm                            {ST(0)       ST(1)   }
  fld qword ptr [C1]           {C1.real             }
  fld qword ptr [C2]           {C2.real     C1.real }
  fsubp ST(1), ST(0)           {result.real         }
  fstp qword ptr [result]      {                    }
  fld qword ptr [C1+8]         {C1.imag             }
  fld qword ptr [C2+8]         {C2.imag     C1.imag }
  fsubp ST(1), ST(0)           {result.imag}
  fstp qword ptr [result+8]
end;
{$else}
inline;
begin
     Result.real:=C1.real-C2.real;
     Result.imag:=C1.imag-C2.imag;
end;
{$endif}

operator * (C1, C2 : TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm                            {ST(0)      ST(1)     ST(2)    ST(3)  }
  fld qword ptr [C2]      {C2.real                              }
  fld qword ptr [C2+8]                  {C2.imag    C2.real                   }
  fld qword ptr [C1]      {C1.real    C2.imag   C2.real         }
  fld qword ptr [C1+8]    {C1.imag    C1.real   C2.imag         C2.real}

  fld ST(3)               {C2.real    C1.imag    C1.real   C2.imag         C2.real}
  fmul ST(0), ST(2)       {C2.r*C1.r  C1.imag    C1.real   C2.imag         C2.real}
  fld ST(1)               {C1.imag    C2.r*C1.r  C1.imag   C1.real   C2.imag         C2.real}
  fmul ST(0), ST(4)       {C1.i*C2.i  C2.r*C1.r  C1.imag   C1.real   C2.imag         C2.real}
  fsubp ST(1), ST(0)      {result.r   C1.imag    C1.real   C2.imag         C2.real}
  fstp qword ptr [result] {C1.imag    C1.real   C2.imag         C2.real}

  fmulp ST(3), ST(0)      {C1.real    C2.imag   C1.imag*C2.real}
  fmulp ST(1), ST(0)      {C1.r*C2.i  C1.i*C2.r}
  faddp St(1), ST(0)      {result.imag}
  fstp qword ptr [result+8]
end;
{$else}
inline;
begin
     Result.real:=C1.real*C2.real-C1.imag*C2.imag;
     Result.imag:=C1.real*C2.imag+C2.real*C1.imag;
end;
{$endif}

operator * (C1: TComplex; r2: Double) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm
   fld qword ptr [r2]
   fld qword ptr [C1]
   fmul ST(0), ST(1)
   fld qword ptr [C1+8]
   fmulp ST(2), ST(0)
   fstp qword [result]
   fstp qword [result+8]
end;
{$else}
inline;
begin
     Result.real:=C1.real*r2;
     Result.imag:=C1.imag*r2;
end;
{$endif}

operator * (r1: Double; C2: TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm
   fld qword ptr [r1]
   fld qword ptr [C2]
   fmul ST(0), ST(1)
   fld qword ptr [C2+8]
   fmulp ST(2), ST(0)
   fstp qword [result]
   fstp qword [result+8]
end;
{$else}
inline;
begin
     Result.real:=C2.real*r1;
     Result.imag:=C2.imag*r1;
end;
{$endif}

operator / (C1, C2 : TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
{$ifdef UseSSE}
asm                   {xmm0         xmm1        xmm2        xmm3      xmm4      }
  movupd xmm0, [C1]   {b1|a1                                                    }
  movupd xmm1, [C2]   {b1|a1        b2|a2                                       }
  movapd xmm2, xmm0   {b1|a1        b2|a2       b1|a1                           }
  movapd xmm3, xmm1   {b1|a1        b2|a2       b1|a1       b2|a2              }

  unpckhpd xmm2, xmm1 {b1|a1        b2|a2       b2|b1       b2|a2 }
  unpcklpd xmm1, xmm0 {b1|a1        a1|a2       b2|b1       b2|a2 }
  mulpd xmm0, xmm3    {b1*b2|a1*a2  a1|a2       b2|b1       b2|a2 }
  mulpd xmm3, xmm3
  mulpd xmm1, xmm2    {b1*b2|a1*a2  a1*b2|a2*b1 b2|b1       b2|a2 }
  haddpd xmm3, xmm3
  hsubpd xmm1, xmm1   {b1*b2|a1*a2  y|y         b2|b1       b2|a2 }
  haddpd xmm0, xmm0   {x|x          y|y         b2|b1       b2|a2 }
  unpcklpd xmm0, xmm1 {y|x          y|y         b2|b1       b2|a2 }
  divpd xmm0, xmm3
  movupd [Result], xmm0
end;
{$else}
asm                       {ST(0)      ST(1)      ST(2)     ST(3)     ST(4)      ST(5)     }
  fld qword ptr [C2]      {C2.real                                                        }
  fld qword ptr [C2+8]    {C2.imag    C2.real                                             }
  fld qword ptr [C1]      {C1.real    C2.imag   C2.real                                   }
  fld qword ptr [C1+8]    {C1.imag    C1.real   C2.imag    C2.real                        }

  fld ST(3)               {C2.real    C1.imag    C1.real   C2.imag   C2.real              }
  fmul ST(0), ST(0)       {C2.real²   C1.imag    C1.real   C2.imag   C2.real              }
  fld ST(3)               {C2.imag    C2.real²   C1.imag   C1.real   C2.imag    C2.real   }
  fmul ST(0), ST(0)       {C2.imag²   C2.real²   C1.imag   C1.real   C2.imag    C2.real   }
  faddp ST(1), ST(0)      {C2.i²+C2.r²C1.imag    C1.real   C2.imag   C2.real              }
  fld Smallest
  faddp ST(1), ST(0)
  fxch ST(4)              {C2.real    C1.imag    C1.real   C2.imag    divisor}

  fld ST(0)               {C2.real    C2.real    C1.imag    C1.real   C2.imag    divisor}
  fmul ST(0), ST(3)       {C2.r*C1.r  C2.real    C1.imag    C1.real   C2.imag    divisor}
  fld ST(2)               {C1.imag    C2.r*C1.r  C2.real   C1.imag    C1.real    C2.imag         divisor}
  fmul ST(0), ST(5)       {C1.i*C2.i  C2.r*C1.r  C2.real   C1.imag    C1.real    C2.imag         divisor}
  faddp ST(1), ST(0)      {a          C2.real    C1.imag    C1.real   C2.imag    divisor}
  fdiv ST(0), ST(5)       {result.r   C2.real    C1.imag    C1.real   C2.imag    divisor}
  fstp qword ptr [Result] {C2.real    C1.imag    C1.real   C2.imag   divisor}

  fmulp ST(1), ST(0)      {C2.r*C1.i  C1.real    C2.imag   divisor}
  fxch ST(2)              {C1.real    C2.imag    C2.r*C1.i divisor}
  fmulp ST(1), ST(0)      {C1.r*C2.i  C2.r*C1.i  divisor}
  fsubp ST(1), ST(0)      {b          divisor}
  fdivrp ST(1), ST(0)     {result.i}
  fstp qword ptr [Result+8]
end;
{$endif}
{$else}
inline;
var
   tmp: Double;
begin
     tmp:=C2.real*C2.real+C2.imag*C2.imag+Smallest;
     Result.real:=(C1.real*C2.real+C1.imag*C2.imag)/tmp;
     Result.imag:=(C2.real*C1.imag-C1.real*C2.imag)/tmp;
end;
{$endif}

operator ** (C1, C2 : TComplex) C : Tcomplex;
{$ifdef UseAsm}
assembler;
assembler;
asm                     {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
  fld qword ptr [C1]    {C1.r                       }
  fld qword ptr [C1+8]  {C1.i   C1.r                       }
  fld LnOneHalf         {ln2    C1.i   C1.r                           }
  fld ST(2)             {C1.r   ln2    C1.i   C1.r                    }
  fmul ST(0),ST(0)      {C1.r^2 ln2    C1.i   C1.r                    }
  fld ST(2)             {C1.i   C1.r^2 ln2    C1.i   C1.r             }
  fmul ST(0),ST(0)      {C1.i^2 C1.r^2 ln2    C1.i   C1.r             }
  fld  Smallest         {Small  C1.i^2 C1.r^2 ln2    C1.i   C1.r             }
  faddp ST(1), ST(0)    {S+i1^2 C1.r^2 ln2    C1.i   C1.r             }
  faddp ST(1), ST(0)    {modC1  ln2    C1.i   C1.r             }
  fyl2x                 {ln(r)  C1.i   C1.r             }
  fxch ST(2)        {C1.r   C1.i      ln(r)}
  fpatan            {argC1  ln(r)                      }

  fld qword ptr [C2]    {C2.r    argC1  ln(r)              }
  fld qword ptr [C2+8]  {C2.i    C2.r   argC1  ln(r)       }
  fld ST(1)             {C2.r    C2.i   C2.r   argC1  ln(r)       }
  fmul ST(0), ST(4)     {r2*lnr  C2.i   C2.r   argC1  ln(r)              }
  fld ST(1)             {C2.i    r2*lnr  C2.i   C2.r   argC1  ln(r) }
  fmul ST(0), ST(4)     {i2*arg  r2*lnr C2.i   C2.r argC1  ln(r)       }
  fsubp ST(1), ST(0)    {a       C2.i   C2.r   argC1  ln(r)              }

  fxch ST(4)            {ln(r)   C2.i   C2.r   argC1  a              }
  fmulp ST(1), ST(0)    {lnr*i2  C2.r   argC1  a              }
  fxch ST(2)            {argC1   C2.r   lnr*i2 a              }
  fmulp ST(1), ST(0)    {r2*arg  lnr*i2 a              }
  faddp ST(1), ST(0)    {b       a}

  fsincos           {cos(b)  sin(b) a}
  fxch ST(1)
  fxch ST(2)        {a cos(b) sin(b) }

  fldl2e            {lg2(e) a      cos(b) sin(b)}
  fmulp st(1),st(0) {a*lg2e cos(b) sin(b)       }
  fld   st(0)       {x      x      cos(b) sin(b)       }
  frndint           {rnd(x) x      cos(b) sin(b)       }
  fsub  st(1),st(0) {rnd(x) frc(x) cos(b) sin(b)       }
  fxch              {frc(x) rnd(x) cos(b) sin(b)       }
  f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
  fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
  faddp ST(1), ST(0)             {2^st   rnd(x) cos(b) sin(b)       }
  fscale            {2^x    rnd(x) cos(b) sin(b)       }
  fstp    st(1)     {exp(a) cos(b) sin(b)              }

  fmul ST(1), ST(0) {exp(a) Res.r  sin(b)              }
  fmulp ST(2),ST(0) {Res.r  Res.i                      }
  fstp  qword ptr [result]  {Res.i                             }
  fstp  qword ptr [result+8]
end;
{$else}
inline;
begin
     Result:=CExp(C2*CLn(C1));
end;
{$endif}

operator ** (C1: TComplex; r2: Double) C : Tcomplex;
{$ifdef UseAsm}
assembler;
asm                     {ST(0)  ST(1)  ST(2)  ST(3)  ST(4) }
  fld qword ptr [C1]    {C1.r                       }
  fld qword ptr [C1+8]  {C1.i   C1.r                       }
  fld LnOneHalf         {ln2    C1.i   C1.r                           }
  fld ST(2)             {C1.r   ln2    C1.i   C1.r                    }
  fmul ST(0),ST(0)      {C1.r^2 ln2    C1.i   C1.r                    }
  fld ST(2)             {C1.i   C1.r^2 ln2    C1.i   C1.r             }
  fmul ST(0),ST(0)      {C1.i^2 C1.r^2 ln2    C1.i   C1.r             }
  fld  Smallest         {Small  C1.i^2 C1.r^2 ln2    C1.i   C1.r             }
  faddp ST(1), ST(0)    {S+i1^2 C1.r^2 ln2    C1.i   C1.r             }
  faddp ST(1), ST(0)    {modC1  ln2    C1.i   C1.r             }
  fyl2x                 {ln(r)  C1.i   C1.r             }
  fxch ST(2)            {C1.r   C1.i      ln(r)}
  fpatan                {argC1  ln(r)                      }

  fld qword ptr [r2]    {C2.r    argC1  ln(r)              }
  fld ST(0)             {C2.r    C2.r   argC1  ln(r)       }
  fmul ST(0), ST(3)     {r2*lnr  C2.r   argC1  ln(r)              }

  fxch ST(3)            {ln(r)   C2.r   argC1  a              }
  fstp ST(0)            {C2.r    argC1  a              }
  fmulp ST(1), ST(0)    {b       a              }

  fsincos           {cos(b)  sin(b) a}
  fxch ST(1)
  fxch ST(2)        {a cos(b) sin(b) }

  fldl2e            {lg2(e) a      cos(b) sin(b)}
  fmulp st(1),st(0) {a*lg2e cos(b) sin(b)       }
  fld   st(0)       {x      x      cos(b) sin(b)       }
  frndint           {rnd(x) x      cos(b) sin(b)       }
  fsub  st(1),st(0) {rnd(x) frc(x) cos(b) sin(b)       }
  fxch              {frc(x) rnd(x) cos(b) sin(b)       }
  f2xm1             {2^st-1 rnd(x) cos(b) sin(b)       }
  fld1              {1      2^st-1 rnd(x) cos(b) sin(b)}
  faddp ST(1), ST(0)             {2^st   rnd(x) cos(b) sin(b)       }
  fscale            {2^x    rnd(x) cos(b) sin(b)       }
  fstp    st(1)     {exp(a) cos(b) sin(b)              }

  fmul ST(1), ST(0) {exp(a) Res.r  sin(b)              }
  fmulp ST(2),ST(0) {Res.r  Res.i                      }
  fstp  qword ptr [result]  {Res.i                             }
  fstp  qword ptr [result+8]
end;
{$else}
inline;
begin
     Result:=CExp(r2*CLn(C1));
end;
{$endif}

function CAbs(C1: TComplex): Double;
{$ifdef UseAsm}
assembler;
asm
  fld qword ptr [C1]
  fmul ST(0),ST(0)
  fld qword ptr [C1+8]
  fmul ST(0),ST(0)
  faddp ST(1), ST(0)
  fsqrt
  fstp qword ptr [Result]
end;
{$else}
inline;
begin
     Result:=sqrt(C1.real*C1.real+C1.imag*C1.imag);
end;
{$endif}

operator = (C1, C2 : TComplex) b : boolean; inline;
begin
  if (C1.real = C2.real) and
    (C1.imag = C2.imag) then Result := true
  else Result := false;
end;

function MakeComplex(r, i: Double): TComplex; inline;
begin
     Result.real:=r;
     Result.imag:=i;
end;

function CModule(C1: TComplex): Double; inline;
begin
     Result:=C1.real*C1.real+C1.imag*C1.imag;
end;

function CArg(C: TComplex): Double;
{$ifdef UseAsm}
assembler;
asm
  fld qword ptr [C+8]
  fld qword ptr [C]
  fpatan
  fstp qword ptr [Result]
end;
{$else}
inline;
begin
     Result:=arctan2(C.imag,C.real);
end;
{$endif}

function CompToStr(C: TComplex; prec: integer = 5): string;
begin
  result := '';

  if abs(C.real)>0 then
  begin
     result := FloatToStrF(C.real, ffGeneral, prec, 0);
     if abs(C.imag)>0 then
       if C.imag > 0 then
        result := '(' + result + '+i*' + FloatToStrF(C.imag, ffGeneral, prec, 0) + ')'
       else result := '(' + result + '-i*' + FloatToStrF(abs(C.imag), ffGeneral, prec, 0) + ')';
  end else if abs(C.imag)>0 then
      result := FloatToStrF(C.imag, ffGeneral, prec, 0) + 'i'
  else Result := '0';
end;


function CSqr(C1: TComplex): TComplex;
{$ifdef UseAsm}
assembler;
{$ifdef UseSSE}
asm
  movddup xmm0, [C1]    {a|a}
  movddup xmm1, [C1+8]  {a|a    b|b}
  movapd xmm2, xmm0     {a|a    b|b     a|a   }
  movapd xmm3, xmm1     {a|a    b|b     a|a   b|b}

  unpcklpd xmm2, xmm1    {a|a    b|b     b|a   b|b}
  unpcklpd xmm3, xmm0    {a|a    b|b     b|a   a|b}
  mulpd xmm0, xmm2      {ab|aa  b|b     b|a   a|b}
  mulpd xmm1, xmm3      {ab|aa  ab|bb   b|a   a|b}
  addsubpd xmm0, xmm1   {2ab|aa-bb  ab|bb   b|a   a|b}

  movupd [Result], xmm0
end;
{$else}
asm
  fld qword ptr [C1]
  fld qword ptr [C1+8]
  fld ST(1)
  fmul ST(0),ST(0)
  fld ST(1)
  fmul ST(0),ST(0)
  fsubp ST(1), ST(0)
  fxch ST(2)

  fmulp ST(1), ST(0)
  fadd  ST(0), ST(0)
  fstp qword ptr [Result+8]
  fstp qword ptr [Result]
end;
{$endif}
{$else}
inline;
begin
     Result.real:=C1.real*C1.real-C1.imag*C1.imag;
     Result.imag:=2*C1.real*C1.imag;
end;
{$endif}

function CLn(C1: TComplex): TComplex;
inline;
begin
     if C1=0 then C1.real:=Smallest;
     Result.real:=ln(CAbs(C1));
     Result.imag:=CArg(C1);
end;

function CExp(C1: TComplex): TComplex;
{$ifdef UseAsm}
assembler;
asm                         {ST(0)  ST(1)  ST(2)}
  fldl2e                    {lg2(e)                            }
  fmul qword ptr [C1]       {a*lg2e                            }
  fld ST(0)                 {x      x                          }
  frndint                   {rnd(x) x                          }
  fsub ST(1), ST(0)         {rnd(x) frc(x)                     }
  fxch                      {frc(x) rnd(x)                     }
  f2xm1                     {2^st-1 rnd(x)                     }
  fld1                      {1      2^st-1 rnd(x)              }
  faddp ST(1), ST(0)        {2^st   rnd(x)                     }
  fscale                    {exp(a) rnd(x)                     }
  fstp ST(1)                {exp(a)                            }

  fld qword ptr [C1+8]      {b       exp(a)                    }
  fsincos                   {cos(b)  sin(b) exp(a)             }

  fmul ST(0), ST(2)         {Ra      sin(b) exp(a)             }
  fstp qword ptr [Result]   {sin(b)  exp(a)                    }
  fmulp ST(1), ST(0)        {Rb                                }
  fstp qword ptr [Result+8]
end;
{$else}
inline;
var
   r: double;
begin
     r:=exp(C1.real);
     Result.real:=r*cos(C1.imag);
     Result.imag:=r*sin(C1.imag);
end;
{$endif}

procedure CSqrAdd(var C1, C2: TComplex);
{$ifdef UseAsm}
assembler;        {C1:=C1^2 + C2}
{$ifdef UseSSE}
asm
  movddup xmm0, [C1]    {a|a}
  movddup xmm1, [C1+8]  {a|a    b|b}
  movapd xmm2, xmm0     {a|a    b|b     a|a   }
  movapd xmm3, xmm1     {a|a    b|b     a|a   b|b}

  unpcklpd xmm2, xmm1    {a|a    b|b     b|a   b|b}
  unpcklpd xmm3, xmm0    {a|a    b|b     b|a   a|b}
  mulpd xmm0, xmm2      {ab|aa  b|b     b|a   a|b}
  mulpd xmm1, xmm3      {ab|aa  ab|bb   b|a   a|b}
  addsubpd xmm0, xmm1   {2ab|aa-bb  ab|bb   b|a   a|b}
  movupd xmm3, [C2]
  addpd xmm0, xmm3

  movupd [C1], xmm0
end;
{$else}
asm
  fld qword ptr [C1]
  fld qword ptr [C1+8]
  fld ST(1)
  fmul ST(0),ST(0)
  fld ST(1)
  fmul ST(0),ST(0)
  fsubp ST(1), ST(0)
  fxch ST(2)

  fmulp ST(1), ST(0)
  fadd  ST(0), ST(0)
  fld qword ptr [C2+8]
  faddp ST(1), ST(0)
  fstp qword ptr [C1+8]
  fld qword ptr [C2]
  faddp ST(1), ST(0)
  fstp qword ptr [C1]
end;
{$endif}
{$else}
inline;
var
   tmp: Double;
begin
     tmp:=C1.real*C1.real-C1.imag*C1.imag+C2.real;
     C1.imag:=2*C1.real*C1.imag+C2.imag;
     C1.real:=tmp;
end;
{$endif}

function CDouble(C1: TComplex): TComplex;
{$ifdef UseAsm}
assembler;
{$ifdef UseSSE}
asm
  movupd xmm0, [C1]
  addpd xmm0, xmm0
  movupd [Result], xmm0
end;
{$else}
asm
  fld qword ptr [C1]
  fld qword ptr [C1+8]
  fadd ST(0), ST(0)
  fstp qword ptr [Result+8]
  fadd ST(0), ST(0)
  fstp qword ptr [Result]
end;
{$endif}
{$else}
inline;
begin
  Result.real := C1.real + C1.real;
  Result.imag := C1.imag + C1.imag;
end;
{$endif}

function CConj(C: TComplex): TComplex;
begin
  Result.real := C.real;
  Result.imag := -C.imag;
end;

end.

