unit UTOrbitTracer;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, LResources, Forms, Controls, Graphics, Dialogs,
  ExtCtrls, StdCtrls, Spin, Grids, UTFractal, Utcomplex, UTRasterizer, UTPixel,
  Menus;

type

  { TOrbitTracer }

  TOrbitTracer = class(TForm)
   but_trace: TButton;
   cmb_tracemode: TComboBox;
   grp_iterations: TGroupBox;
   grp_Point: TGroupBox;
   lbl_pointR: TLabel;
   lbl_pointI: TLabel;
   grd_menu_savegrd: TMenuItem;
   grd_menu: TPopupMenu;
   dlg_tracersave: TSaveDialog;
   sed_PointR: TFloatSpinEdit;
   sed_PointI: TFloatSpinEdit;
   grp_tracing: TGroupBox;
   Image1: TImage;
   img_main: TImage;
   sed_iterations: TSpinEdit;
   grd_output: TStringGrid;
   procedure but_traceClick(Sender: TObject);
   procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
   procedure FormCreate(Sender: TObject);
   procedure grd_menu_savegrdClick(Sender: TObject);
  private
  procedure FillGrid;
   TraceData: TTraceData;
  public
    { public declarations }
    procedure UpdateCanvas;
  end;

type
  TOrbitRasterizer = class(TRasterizer)
  protected
    procedure Execute; override;
  end;

var
  OrbitTracer: TOrbitTracer;
  img_main_Bitmap: ^LongWord;
  OTFractal: TFractal;

implementation

uses UTFractalGenerator;

{ TOrbitTracer }

procedure TOrbitTracer.FormClose(Sender: TObject; var CloseAction: TCloseAction
 );
begin
 if Assigned(OTFractal) then OTFractal.Free;
 if Assigned(img_main_Bitmap) then System.FreeMem(img_main_Bitmap, 250*250*4);
end;

procedure TOrbitTracer.FormCreate(Sender: TObject);
begin
  OTFractal := TFractal.Create(Fractal);
  OTFractal.RasterizerThread := TOrbitRasterizer;
  OTFractal.SetSize(MakeComplex(-2.5, 2.5), MakeComplex(2.5, -2.5), 250, 250);
  OTFractal.Iterations := 10;
  //if not Assigned(img_main_Bitmap) then
  GetMem(img_main_Bitmap, 250*250*4);
end;

procedure TOrbitTracer.grd_menu_savegrdClick(Sender: TObject);
var
  f, s: string;
  i: integer;
  D: TextFile;
begin
  if dlg_tracersave.Execute then
  begin
    f := dlg_tracersave.FileName;
    System.Assign(D, f);
    System.Rewrite(D);

    for i := 1 to TraceData.iterations-1 do
    begin
      s := IntToStr(i) + ' ';
      s += FloatToStr(TraceData.Z[i].real) + ' ';
      s += FloatToStr(TraceData.Z[i].imag) + ' ';
      if cmb_tracemode.ItemIndex = 0 then
        s += FloatToStr(TraceData.LyaPartSums[i])
      else s += FloatToStr(TraceData.LyaSum[i]);
      WriteLn(D, s);
    end;
    System.Close(D);
  end;
end;

procedure TOrbitTracer.FillGrid;
var
  i: integer;
begin
  with grd_output do
  begin
    ColCount := 4;
    Cells[0,0] := 'Nr.';
    Cells[1,0] := 'Realteil';
    Cells[2,0] := 'Imaginärteil';
    Cells[3,0] := 'Lyapunov-Summe';

    RowCount := TraceData.iterations;
    for i := 1 to RowCount-1 do
    begin
      Cells[0,i] := IntToStr(i);
      Cells[1,i] := FloatToStr(TraceData.Z[i].real);
      Cells[2,i] := FloatToStr(TraceData.Z[i].imag);
      if cmb_tracemode.ItemIndex = 0 then
        Cells[3,i] := FloatToStr(TraceData.LyaPartSums[i])
      else Cells[3,i] := FloatToStr(TraceData.LyaSum[i])
    end;
    AutoSizeColumns;
  end;
end;

procedure TOrbitTracer.but_traceClick(Sender: TObject);
begin
  OTFractal.Draw;
  UpdateCanvas;
  with TraceData do
  begin
    iterations := sed_iterations.Value;
    C := MakeComplex( sed_PointR.Value , sed_PointI.Value);
  end;
  OTFractal.TraceOrbit(TraceData);
  FillGrid;
end;

procedure TOrbitTracer.UpdateCanvas;
var
  X,Y: integer;
begin
  OrbitTracer.img_main.Canvas.Clear;
  for Y:=0 to OTFractal.Height_pixels - 1 do
    for X:=0 to OTFractal.Width_pixels - 1 do
      OrbitTracer.img_main.Canvas.Pixels[X,Y] := PixelToColor(img_main_Bitmap[X+Y*250]);
end;

procedure TOrbitRasterizer.Execute;
var
  X,Y: integer;
begin
  for Y:=0 to OTFractal.Height_pixels - 1 do
    for X:=0 to OTFractal.Width_pixels - 1 do
      img_main_Bitmap[X+Y*250] := ColorFunction(X,Y);
end;

initialization
  {$I utorbittracer.lrs}

end.

