unit UTPixel;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, Math, FPImage;

const
     ClBlack32 = $FF000000;
     ClRed32 = $FF0000FF;
     ClGreen32 = $FF00FF00;
     ClBlue32 = $FFFF0000;
     ClWhite32 = $FFFFFFFF;
     TexBackground = $FFECEBEA;
     //Hintergrund für Texturen

type
    TPixel = Longword;
    PPixel = ^Longword;
    {OpenGl ABGR Format}

function FPColorToPixel(F: TFPColor): TPixel;
function ColorToPixel(C: TColor): TPixel;
function PixelToColor(P: TPixel): TColor;
function RGBToPixel(R,G,B: Byte): TPixel;
function HSLToPixel(h,s,l: Single): TPixel;
function Gray32(P: TPixel): TPixel; overload;
function Gray32(P: TPixel; alpha: single): TPixel; overload;
function Blacken(P: TPixel; a: byte): TPixel;
function BlendPixel(A,B: TPixel; r: Double): TPixel;

implementation

function FPColorToPixel(F: TFPColor): TPixel;
begin
  Result := $FF000000 or ((F.Blue shr 8) and $FF) or (F.Green and $FF00) or ((F.Red shl 8) and $FF0000);
end;

function ColorToPixel(C: TColor): TPixel;
begin
  Result := C or $FF000000;
end;

function PixelToColor(P: TPixel): TColor;
begin
  Result := P and $FFFFFF;
end;

function RGBtoPixel(R,G,B: Byte): TPixel; inline;
begin
     Result:=$FF000000 or (B shl 16) or (G shl 8) or R;
end;

function HSLToPixel(h,s,l: Single): TPixel;
const
  OneOverThree = 1 / 3;
var
  M1, M2: Single;
  R, G, B: Byte;

  function HueToColor(Hue: Single): Byte;
  var
    V: Double;
  begin
    Hue := Hue - floor(Hue);
    if 6 * Hue < 1 then V := M1 + (M2 - M1) * Hue * 6
    else if 2 * Hue < 1 then V := M2
    else if 3 * Hue < 2 then V := M1 + (M2 - M1) * (2 / 3 - Hue) * 6
    else V := M1;
    Result := Round(255 * V);
  end;

begin
  if S = 0 then
  begin
    R := Round(255 * L);
    G := R;
    B := R;
  end
  else
  begin
    if L <= 0.5 then M2 := L * (1 + S)
    else M2 := L + S - L * S;
    M1 := 2 * L - M2;
    R := HueToColor(H + OneOverThree);
    G := HueToColor(H);
    B := HueToColor(H - OneOverThree)
  end;
  Result := RGBToPixel(R, G, B);
end;

function BlendPixel(A,B: TPixel; r: Double): TPixel;
var
   R1,R2,G1,G2,B1,B2: Byte;
begin
     B1:=(A and $FF0000) shr 16;
     B2:=(B and $FF0000) shr 16;
     G1:=(A and $FF00) shr 8;
     G2:=(B and $FF00) shr 8;
     R1:= A and $FF;
     R2:= B and $FF;

     R1:=byte(round(R1*r+R2*(1-r)));
     G1:=byte(round(G1*r+G2*(1-r)));
     B1:=byte(round(B1*r+B2*(1-r)));

     Result:=RGBToPixel(R1,G1,B1);
end;

function Gray32(P: TPixel): TPixel; overload;
var
   A, R, G, B: Byte;
begin
     R:=(P and $FF0000) shr 16;
     G:=(P and $FF00) shr 8;
     B:= P and $FF;
     A:=(R+G+B) div 3;
     Result:=$FF000000 or (A shl 16) or (A shl 8) or A;
end;

function Gray32(P: TPixel; alpha: single): TPixel; overload;
var
   A, R, G, B: Byte;
begin
     R:=(P and $FF0000) shr 16;
     G:=(P and $FF00) shr 8;
     B:= P and $FF;
     A:=byte(round(((R+G+B) div 3) * alpha));
     Result:=$FF000000 or (A shl 16) or (A shl 8) or A;
end;

function Blacken(P: TPixel; a: byte): TPixel;
var
   R, G, B: Byte;
begin
     R:=(P and $FF0000) shr 16;
     G:=(P and $FF00) shr 8;
     B:= P and $FF;
     R:=(R + a) div 2;
     G:=(G + a) div 2;
     B:=(B + a) div 2;
     Result:=$FF000000 or (R shl 16) or (G shl 8) or B;
end;

end.

