unit UTRasterizer;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, UTPixel, UOGL, math
  {$ifdef unix}
    ,unix
  {$endif}
  {$ifdef win32}
    ,windows
  {$endif}
  ;

type TColorFunction = function(X, Y: integer): TPixel of object; register;

Type
    TRasterizer = class(TThread)
    public
       Mode: Boolean;
       ColorFunction: TColorFunction;
       Constructor Create(CreateSuspended : boolean);
    end;

type TRasterizerClass = class of TRasterizer;

Type
    TOGLRasterizer = class(TRasterizer)
    private
      procedure ShowStatus;
      procedure StopTime;
      procedure StartTime;
      procedure PrepareDS;
    protected
      procedure Execute; override;
    end;

      //procedure DrawCoordinates;
var
  Tex: OGLTiledTexture;
  DrawCoordSystem: boolean;
  CSXWidth, CSXHeight, CSXTop: integer;
  CSYWidth, CSYHeight, CSYLeft: integer;
  CSFontSize, CSLineSize: integer;
  {$ifdef unix} time1,time2: timeval; {$endif}
  {$ifdef win32} freq,time1,time2: int64; {$endif}

implementation

uses UTFractalGenerator;

constructor TRasterizer.Create(CreateSuspended : boolean);
begin
  FreeOnTerminate := True;
  Mode:=False;
  inherited Create(CreateSuspended);
  DrawCoordSystem := false;
end;

procedure TOGLRasterizer.ShowStatus;
begin
  UpdateTexture(Tex);
  FG.img_main.Invalidate;
end;

procedure TOGLRasterizer.PrepareDS;
begin
  //Zeichenfläche initialisieren
  if Fractal.SizeChanged then
  begin
    //FG.Caption:=FloatToStr(ceil(-2.5));
    FG.scr_Height:=Fractal.Height_pixels;
    FG.scr_Width:=Fractal.Width_pixels;
    if DrawCoordSystem then
    begin
      CSFontSize := round(sqrt(Fractal.Width_pixels+Fractal.Height_pixels)/1.25);
      CSLineSize := round(CSFontSize/10);
      CSYWidth := CSFontSize*5; //round(Fractal.Width_pixels * 0.1);
      CSYHeight := Fractal.Height_pixels;
      CSYLeft := Fractal.Width_pixels+1;
      CSXHeight := CSFontSize*3; //round(Fractal.Height_pixels * 0.05);
      CSXWidth := Fractal.Width_pixels;
      CSXTop := Fractal.Height_pixels+1;
      GenerateOGLTiledTexture(Tex,Fractal.Width_pixels+CSYWidth,
        Fractal.Height_pixels+CSXHeight);
      FG.scr_Height+=CSXHeight+1;
      FG.scr_Width+=CSYWidth+1;
    end else
          GenerateOGLTiledTexture(Tex,Fractal.Width_pixels,
        Fractal.Height_pixels);
    FG.MeasureScrollBars;
    Fractal.SizeChanged:=false;
  end;
end;

procedure TOGLRasterizer.Execute;
var
  i, k: integer;
  cl: TPixel;
  run_nmbr: integer;
  QWidth, QHalfWidth: integer;
  X,Y, UpdateLines: integer;
  label ende;
begin
  Synchronize(@PrepareDS);
  StartTime;
  if Mode then
  begin
       for i:=0 to Fractal.Height_pixels-1 do
       begin
          if (i+1) mod 500 = 0 then Synchronize(@Showstatus);
          if Terminated then goto Ende;
          for k:=0 to Fractal.Width_pixels-1 do
          begin
            cl:=ColorFunction(k,i);
            uogl.SetPixel(Tex, TPixel(cl), k, i);
          end;
       end;
       Synchronize(@Showstatus);
  end else begin

    for run_nmbr:=1 to 3 do
    begin
      UpdateLines:=(Fractal.Height_pixels-1) div (run_nmbr+1);
      QWidth := 1;
      for i:=2 downto run_nmbr do QWidth := QWidth * 3;
      QHalfWidth:=trunc( QWidth / 2);
      Y := QHalfWidth;
      Synchronize(@ShowStatus);
      while Y <= Fractal.Height_pixels - QWidth do
      begin
           X := QHalfWidth;
           while X <= Fractal.Width_pixels - QWidth do
           begin
                if Terminated then goto Ende;
                if Fractal.calculated[X,Y] = false then begin
                     cl:=ColorFunction(X,Y);
                     FillQuad(Tex, TPixel(cl), X-QHalfWidth, Y-QHalfWidth, QWidth);
                end;
                inc(X, QWidth);
           end;
           inc(Y, QWidth);
           if Y mod UpdateLines = 0 then Synchronize(@ShowStatus);
      end;
    end;
  end;
ende:
  Synchronize(@Showstatus);
  Synchronize(@StopTime);
  Fractal.Validate;
end;

procedure TOGLRasterizer.StopTime;
begin
{$ifdef unix}
  fpgettimeofday(@time2, nil);
  time2.tv_sec -= time1.tv_sec;
  time2.tv_usec -= time1.tv_usec;
  FG.stb_main.panels[1].text := FloatToStr( time2.tv_sec + time2.tv_usec / 1000000);
{$endif}
{$ifdef win32}
  QueryPerformanceCounter(time2);
  time1 := time2 - time1;
  FG.stb_main.panels[1].text := FloatToStr( time1 / freq);
{$endif}
end;

procedure TOGLRasterizer.StartTime;
begin
{$ifdef unix}
  fpgettimeofday(@time1, nil);
{$endif}
{$ifdef win32}
  QueryPerformanceFrequency(freq);
  QueryPerformanceCounter(time1);
{$endif}
end;

end.

